@@ -12,13 +12,18 @@ import urllib2
 import cgi
 import re
 
-__version__  = '0.9.3'
+__version__  = '0.10.1'
 __author__ = "Hiroki Ohtani (liris.pp@gmail.com)"
 __contributors__ = ["Junji NAKANISHI <daemonfreaks@gmail.com>"]
 __copyright__ = "Copyright (c) 2006 Hiroki Ohtani"
 __license__ = "LGPL"
 
 
+COMPATIBLE_WITH_09 = False
+
+class HyperestraierError(Exception):
+    '''Base exception for Hyperestraier client.'''
+
 class Document(object):
     '''Document class for hyperestrair.
     '''
@@ -340,13 +345,26 @@ class NodeResult(object):
         """
         Get the value of hint information.
         `key' specifies the key of a hint. 
-        "VERSION", "NODE", "HIT", "HINT#n", "DOCNUM", "WORDNUM",
-        "TIME", "TIME#n", "LINK#n", and "VIEW" are provided for keys.
-        The return value is the hint or `nil' if the key does not exist.
+        'VERSION', 'NODE', 'HIT', 'HINT#n', 'DOCNUM', 'WORDNUM',
+        'TIME', 'TIME#n', 'LINK#n', and 'VIEW' are provided for keys.
+        The return value is the hint or None if the key does not exist.
         """
         return self.hints[key]
 
 
+def _success(result):
+    return True
+
+def _err(result):
+    if not COMPATIBLE_WITH_09:
+        raise HyperestraierError(result)
+    else:
+        return False
+
+
+class UrlNotSpecifiedException(Exception):
+    pass
+
 class Transport(object):
     def __init__(self):
         self.url = None
@@ -354,20 +372,35 @@ class Transport(object):
         self.pxport = 0
         self.timeout = 0
         self.auth = None
-    
-    def send(self, command, headers = {}, body=None, autoContentType = True):
+
+    def extract(self, callback):
+        return callback(None)
+
+    def sendAndExtract(self, command, callback, errback, extractCallback,
+                       headers = {}, body=None, autoContentType = True):
+        result = self.send(command, callback, errback,
+                           headers, body, autoContentType)
+        return extractCallback(result)
+
+    def send(self, command, callback, errback,
+             headers = {}, body=None, autoContentType = True):
         # return self._send(command, headers, body)
         try:
-            return self._send(command, headers, body)
+            status, result = self._send(command, headers, body)
+            return callback(result)
         except urllib2.HTTPError, e:
-            return e.code, str(e)
+            status, result = e.code, str(e)
         except Exception, e:
-            return -1, str(e)
+            status, result =  -1, str(e)
+        return errback(result)
+        
 
     def _send(self, command, headers = {}, body=None, autoContentType=True):
         if not self.url:
-            return -1, None
+            raise UrlNotSpecifiedException
+
         url = self.url + "/" + command
+        headers = headers.copy()
         if not headers.has_key("Content-Type") and autoContentType:
             headers["Content-Type"] = "application/x-www-form-urlencoded"
         if self.auth:
@@ -390,12 +423,78 @@ class Transport(object):
         
         return code, result.decode("utf-8")
 
+try:
+    from twisted.internet import defer, reactor
+    from twisted.web.client import getPage
+    
+    def unicodeFromUTF8(result):
+        return result.decode("utf-8")
+
+    class AsynTransport(object):
+        """
+        Asynchornous hyperestraier support.
+        If you use this transport,
+        All Node API can run asynchonously and always return deferred object.
+        """
+        def __init__(self):
+            self.url = None
+            self.pxhost = None
+            self.pxport = 0
+            self.timeout = 0
+            self.auth = None
+            
+        def extract(self, callback):
+            deferred = defer.Deferred()
+            def _f():
+                deferred.callback(callback(None))
+
+            reactor.callLater(0, _f)
+            return deferred
+
+        def sendAndExtract(self, command, callback, errback, extractCallback,
+                 headers = {}, body=None, autoContentType = True):
+            deferred = self.send(command, callback, errback,
+                               headers, body, autoContentType)
+            return deferred.addCallback(extractCallback)
+    
+        def send(self, command, callback, errback,
+                 headers = {}, body=None, autoContentType = True):
+            if not self.url:
+                raise UrlNotSpecifiedException
+            
+            headers = headers.copy()
+            url = self.url + "/" + command
+            if not headers.has_key("Content-Type") and autoContentType:
+                headers["Content-Type"] = "application/x-www-form-urlencoded"
+            if self.auth:
+                encodedAuth = base64.encodestring(self.auth).replace("\n", "")
+                headers["Authorization"] = "Basic " + encodedAuth
+            if body:
+                headers["Content-Length"] = "%d" % len(body)
+            headers["User-Agent"] = "HyperEstraierForPython/1.0.0"
+
+            deferred = getPage(url, method = (body and "POST" or "GET"),
+                               headers = headers, postdata = body)
+            deferred.addCallback(
+                unicodeFromUTF8).addCallback(
+                callback).addErrback(
+                errback)
+            return deferred
+            
+            
+  
+except:
+    pass
+
 class Node(object):
     """
     Abstraction of connection to P2P node.
     """
-    def __init__(self):
-        self.transport = Transport()
+    def __init__(self, transport = None):
+        if transport:
+            self.transport = transport
+        else:
+            self.transport = Transport()
         self.name = None
         self.label = None
         self.dnum = -1
@@ -444,18 +543,16 @@ class Node(object):
         Synchronize updating contents of the database.
         The return value is True if success, else it is False.
         """
-        status, result = self.transport.send("sync")
-        
-        return status == 200
+        return  self.transport.send("sync", _success, _err)
+
 
     def optimize(self):
         """
         Optimize the database.
         The return value is True if success, else it is False.
         """
-        status, result = self.transport.send("optimize")
-        
-        return status == 200
+        return self.transport.send("optimize", _success, _err)
+
 
     def put_doc(self, doc):
         """
@@ -464,12 +561,11 @@ class Node(object):
         The document object should have the URI attribute.
         The return value is True if success, else it is False.
         """
-        status, result = self.transport.send(
+        return  self.transport.send(
             "put_doc",
+            _success, _err,
             {"Content-Type":"text/x-estraier-draft"},
             doc.dump_draft())
-        
-        return status == 200
 
     def out_doc(self, id):
         """
@@ -477,11 +573,10 @@ class Node(object):
         `id' specifies the ID number of a registered document.
         The return value is True if success, else it is False.
         """
-        status, result = self.transport.send(
+        return self.transport.send(
             "out_doc",
+            _success, _err,
             body = "id=" + str(id))
-        
-        return status == 200
 
     def out_doc_by_uri(self, uri):
         """
@@ -490,11 +585,10 @@ class Node(object):
         The return value is True if success, else it is False.
         """
         body = "uri=" + _escape(uri)
-        status, result = self.transport.send(
+        return self.transport.send(
             "out_doc",
+            _success, _err,
             body = body)
-        
-        return status == 200
 
     def edit_doc(self, doc):
         """
@@ -502,12 +596,11 @@ class Node(object):
         `doc' specifies a document object.
         The return value is True if success, else it is False.
         """
-        status, result = self.transport.send(
+        return self.transport.send(
             "edit_doc",
+            _success, _err,
             {"Content-Type":"text/x-estraier-draft"},
             doc.dump_draft())
-        
-        return status == 200
 
     def get_doc(self, id):
         """
@@ -516,12 +609,14 @@ class Node(object):
         The return value is a document object.  On error, None is returned.
         """
         body = "id=" + str(id)
-        status, result = self.transport.send("get_doc", body = body)
-        
-        if status == 200:
+        def success(result):
             return Document(result)
-        return None
+        
+        def err(result):
+            return None
 
+        return self.transport.send("get_doc", success, err, body = body)
+        
     def get_doc_by_uri(self, uri):
         """
         Retrieve a document.
@@ -529,11 +624,13 @@ class Node(object):
         The return value is a document object.  On error, None is returned.
         """
         body = "uri=" + _escape(uri)
-        status, result = self.transport.send("get_doc", body = body)
-        
-        if status == 200:
+        def success(result):
             return Document(result)
-        return None
+        
+        def err(result):
+            return None
+
+        return  self.transport.send("get_doc", success, err, body = body)
 
     def get_doc_attr(self, id, name):
         """
@@ -544,11 +641,13 @@ class Node(object):
         or None if it does not exist.
         """
         body = "id=" + str(id) + "&attr=" + _escape(name)
-        status, result = self.transport.send("get_doc_attr", body = body)
-        
-        if status == 200:
+        def success(result):
             return result.strip()
-        return None
+        
+        def err(result):
+            return None
+
+        return self.transport.send("get_doc_attr", success, err, body = body)        
 
     def get_doc_attr_by_uri(self, uri, name):
         """
@@ -559,11 +658,13 @@ class Node(object):
         or None if it does not exist.
         """
         body = "uri=" + _escape(uri) + "&attr=" + _escape(name)
-        status, result = self.transport.send("get_doc_attr", body = body)
-        
-        if status == 200:
+        def success(result):
             return result.strip()
-        return None
+        
+        def err(result):
+            return None
+        
+        return self.transport.send("get_doc_attr", success, err, body = body)
 
     def etch_doc(self, id):
         """
@@ -573,17 +674,19 @@ class Node(object):
         or None on error.
         """
         body = "id=" + str(id)
-        status, result = self.transport.send("etch_doc", body = body)
-        
-        if status == 200:
+        def success(result):
             kwords = {}
             for line in result.splitlines():
                 pair = line.split("\t")
                 if len(pair) > 1:
                     kwords[pair[0]] = pair[1]
             return kwords
-        return None
-
+        
+        def err(result):
+            return None
+        
+        return self.transport.send("etch_doc", success, err, body = body)
+        
     def etch_doc_by_uri(self, uri):
         """
         Extract keywords of a document specified by URI.
@@ -592,17 +695,19 @@ class Node(object):
         or None on error.
         """
         body = "uri=" + _escape(uri)
-        status, result = self.transport.send("etch_doc", body = body)
-        
-        if status == 200:
+        def success(result):
             kwords = {}
             for line in result.splitlines():
                 pair = line.split("\t")
                 if len(pair) > 1:
                     kwords[pair[0]] = pair[1]
             return kwords
-        return None
-
+        
+        def err(result):
+            return None
+        
+        return self.transport.send("etch_doc", success, err, body = body)
+        
     def uri_to_id(self, uri):
         """
         Get the ID of a document specified by URI.
@@ -610,56 +715,73 @@ class Node(object):
         The return value is the ID of the document.  On error, -1 is returned.
         """
         body = "uri=" + _escape(uri)
-        status, result = self.transport.send("uri_to_id", body = body)
-        
-        if status == 200:
+        def success(result):
             return int(result)
-        return None
+        def err(result):
+            return None
+
+        return  self.transport.send("uri_to_id", success, err, body = body)
     
     def get_name(self):
         """
         Get the name.
         The return value is the name.  On error, None is returned.
         """
+        def get(result):
+            return self.name
+        
         if self.name == None:
-            self.set_info()
-        return self.name
+            return self._set_info(get)
+        #return self.name
+        return self.transport.extract(get)
     
     def get_label(self):
         """
         Get the label.
         The return value is the label.  On error, None is returned.
         """
+        def get(result):
+            return self.label
+
         if self.label == None:
-            self.set_info()
-        return self.label
+            self._set_info(get)
+        return self.transport.extract(get)
 
     def get_doc_num(self):
         """
         Get the number of documents.
         The return value is the number of documents.  On error, -1 is returned.
         """
+        def get(result):
+            return self.dnum
+
         if self.dnum < 0:
-            self.set_info()
-        return self.dnum
+            return self._set_info(get)
+        return self.transport.extract(get)
 
     def get_word_num(self):
         """
         Get the number of unique words.
         The return value is the number of unique words.  On error, -1 is returned.
         """
+        def get(result):
+            return self.wnum
+
         if self.wnum < 0:
-            self.set_info()
-        return self.wnum
+            return self._set_info(get)
+        return self.transport.extract(get)
 
     def get_size(self):
         """
         Get the size of the datbase.
         The return value is the size of the datbase.  On error, -1.0 is returned.
         """
+        def get(result):
+            return self.size
+
         if self.size < 0:
-            self.set_info()
-        return self.size
+            return self._set_info(get)
+        return self.transport.extract(get)
 
     def get_cache_usage(self):
         """
@@ -667,11 +789,15 @@ class Node(object):
         The return value is the usage ratio of the cache. 
         On error, -1.0 is returned.
         """
-        status, result = self.transport.send("cacheusage",
-                                             autoContentType = False)
-        if status != 200 or not result:
+        def err(result):
             return -1.0
-        return float(result)
+
+        def success(result):
+            return float(result)
+
+        return self.transport.send("cacheusage",
+                                   success, err,
+                                   autoContentType = False)
 
     def get_admins(self):
         """
@@ -679,9 +805,12 @@ class Node(object):
         The return value is an array object of names of administrators.
         On error, None is returned.
         """
-        if self.admins == None:
-            self.set_info()
-        return self.admins
+        def get(result):
+            return self.admins
+
+        if self.admins < 0:
+            return self._set_info(get)
+        return self.transport.extract(get)
 
     def get_users(self):
         """
@@ -689,9 +818,12 @@ class Node(object):
         The return value is an array object of names of users.
         On error, None is returned.
         """
-        if self.users == None:
-            self.set_info()
-        return self.users
+        def get(result):
+            return self.users
+
+        if self.users < 0:
+            return self._set_info(get)
+        return self.transport.extract(get)
 
     def get_links(self):
         """
@@ -701,9 +833,12 @@ class Node(object):
         the URL, the label, and the score.  
         On error, None is returned.
         """
-        if self.links == None:
-            self.set_info()
-        return self.links
+        def get(result):
+            return self.links
+
+        if self.links < 0:
+            return self._set_info(get)
+        return self.transport.extract(get)
 
     def search(self, condition, depth = 0):
         """
@@ -712,64 +847,67 @@ class Node(object):
         `depth' specifies the depth of meta search.
         The return value is a node result object.  On error, None is returned.
         """
+        def success(result):
+            lines = result.splitlines()
+            docs = []
+            hints = {}
+            nres = NodeResult(docs, hints)
+            border = lines.pop(0)
+            isend = False
+            while len(lines):
+                line = lines.pop(0)
+                if line.startswith(border):
+                    if line[len(border)+1:] == ":END":
+                        isend = True
+                    lines.insert(0, line)
+                    break
+                elems = line.split("\t", 1)
+                if len(elems) == 2:
+                    hints[elems[0]] = elems[1]
+
+            while not isend and len(lines):
+                line = lines.pop(0)
+                if line.startswith(border):
+                    rdattrs = {}
+                    sb = []
+                    rdvector = ""
+                    while len(lines):
+                        rdline = lines.pop(0).strip()
+                        if not rdline:
+                            break
+                        if rdline.startswith("%"):
+                            vecelem = rdline.split("\t", 1)
+                            if len(vecelem) == 2 and vecelem[0] == "%VECTOR":
+                                rdvector = vecelem[1]
+                        else:
+                            elems = rdline.split("=", 1)
+                            if len(elems)==2:
+                                rdattrs[elems[0]] = elems[1]
+                    while len(lines):
+                        rdline = lines.pop(0)
+                        if not rdline:
+                            break
+                        sb.append(rdline)
+                    rduri = rdattrs.setdefault("@uri", None)
+                    rdsnippet = "\n".join(sb)
+                    if rduri:
+                        rdoc = ResultDocument(rduri, rdattrs, rdsnippet, rdvector)
+                        docs.append(rdoc)
+                if line.endswith(":END"):
+                    isend = True
+                    break
+
+            if isend:
+                return nres
+            return None
+
+        def err(resutl):
+            return None
+        
         body = _condToQuery(condition, depth,
                                 self.wwidth, self.hwidth, self.awidth)
-        status, result = self.transport.send("search", body=body)
-        if status != 200 or not result:
-            return None
-        lines = result.splitlines()
-        docs = []
-        hints = {}
-        nres = NodeResult(docs, hints)
-        border = lines.pop(0)
-        isend = False
-        while len(lines):
-            line = lines.pop(0)
-            if line.startswith(border):
-                if line[len(border)+1:] == ":END":
-                    isend = True
-                lines.insert(0, line)
-                break
-            elems = line.split("\t", 1)
-            if len(elems) == 2:
-                hints[elems[0]] = elems[1]
+        return self.transport.send("search", success, err, body=body)
 
-        while not isend and len(lines):
-            line = lines.pop(0)
-            if line.startswith(border):
-                rdattrs = {}
-                sb = []
-                rdvector = ""
-                while len(lines):
-                    rdline = lines.pop(0).strip()
-                    if not rdline:
-                        break
-                    if rdline.startswith("%"):
-                        vecelem = rdline.split("\t", 1)
-                        if len(vecelem) == 2 and vecelem[0] == "%VECTOR":
-                            rdvector = vecelem[1]
-                    else:
-                        elems = rdline.split("=", 1)
-                        if len(elems)==2:
-                            rdattrs[elems[0]] = elems[1]
-                while len(lines):
-                    rdline = lines.pop(0)
-                    if not rdline:
-                        break
-                    sb.append(rdline)
-                rduri = rdattrs.setdefault("@uri", None)
-                rdsnippet = "\n".join(sb)
-                if rduri:
-                    rdoc = ResultDocument(rduri, rdattrs, rdsnippet, rdvector)
-                    docs.append(rdoc)
-            if line.endswith(":END"):
-                isend = True
-                break
-                
-
-        if isend:
-            return nres
-        return None
 
     def set_snippet_width(self, wwidth, hwidth, awidth):
         """
@@ -802,8 +940,7 @@ class Node(object):
         The return value is true if success, else it is false.
         """
         body = "name=" + _escape(name) + "&mode=" + str(mode)
-        status, result = self.transport.send("_set_user", body=body)
-        return status == 200
+        return self.transport.send("_set_user", _success, _err, body=body)
 
     def set_link(self, url, label, credit):
         """
@@ -817,40 +954,47 @@ class Node(object):
         body = "url=" + _escape(url) + "&label=" + label
         if credit >= 0:
             body += "&credit=" + str(credit)
-        status, result = self.transport.send("_set_link", body=body)
-        return status == 200
-
-    def set_info(self):
-        status, result = self.transport.send("inform",
-                                                  autoContentType = False)
-        if status != 200 or not result:
-            return
-        lines = result.splitlines()
-        line = lines.pop(0)
-        elems = line.split("\t")
-        if len(elems) != 5:
-            return
-        self.name = elems[0]
-        self.label = elems[1]
-        self.dnum = int(elems[2])
-        self.wnum = int(elems[2])
-        self.size = int(elems[2])
         
+        return self.transport.send("_set_link", _success, _err, body=body)
 
-        self.admins = []
-        self.users = []
-        self.links = []
-        currentScope = (self.admins, lambda x: x)
-        remainQueue = [(self.users, lambda x: x),
-                       (self.links, lambda x: x.split("\t", 3))]
-        for line in lines:
-            if not line:
-                if len(remainQueue):
-                    currentScope = remainQueue.pop(0)
+    def _set_info(self, callback):
+        def err(result):
+            pass
+
+        def success(result):
+            lines = result.splitlines()
+            line = lines.pop(0)
+            elems = line.split("\t")
+            if len(elems) != 5:
+                return
+            self.name = elems[0]
+            self.label = elems[1]
+            self.dnum = int(elems[2])
+            self.wnum = int(elems[2])
+            self.size = int(elems[2])
+
+
+            self.admins = []
+            self.users = []
+            self.links = []
+            currentScope = (self.admins, lambda x: x)
+            remainQueue = [(self.users, lambda x: x),
+                           (self.links, lambda x: x.split("\t", 3))]
+            for line in lines:
+                if not line:
+                    if len(remainQueue):
+                        currentScope = remainQueue.pop(0)
+                    else:
+                        break
                 else:
-                    break
-            else:
-                currentScope[0].append(currentScope[1](line))
+                    currentScope[0].append(currentScope[1](line))
+            
+            return self
+
+
+        return  self.transport.sendAndExtract("inform",
+                                              success, err, callback,
+                                              autoContentType = False)
 
 
 SNIP_PATTERN = re.compile(u"[ \t\r\n\v\f]+")
